<?php
/*
  サンプルプログラム
  Copyright (C) 2012 Masanori OKIBAYASHI all rights reserved.
*/


/******************************

     Web APIからのアクセスで
     行われる処理(サービス)

 ******************************/


// データベース制御クラス - スーパークラス
// (PDOによるアクセス、プリペアードSQLの実行)
abstract class DB {

  protected $db, $lastInsertId;   // データベース、データ挿入時のID

  abstract protected function initDb(); // データベース初期設定

  // 参照系(SELECT)の実行
  // (戻り値は複数件なら連想配列の配列、1件のみなら連想配列)
  protected function get($sql, $args = null) {
    $stmt = $this->exec($sql, $args);
    $results = $stmt->fetchAll(PDO::FETCH_ASSOC);
    if (count($results) == 1)  $results = $results[0];
    unset($stmt);
    return $results;
  }

  // 更新系(INSERT, UPDATE, DELETE)の実行
  // (戻り値は更新された行数)
  protected function set($sql, $args = array()) {
    $stmt = $this->exec($sql, $args);
    $rows = $stmt->rowCount();
    unset($stmt);
    return $rows;
  }

  // SQL文の実行
  // (戻り値はStatement、同時に挿入時IDも設定)
  protected function exec($sql, $args) {
    $this->lastInsertId = null;
    $stmt = $this->db->prepare($sql);
    $stmt->execute($args);
    $this->lastInsertId = $this->db->lastInsertId() + 0;
    return $stmt;
  }

  // コンストラクタ(DBオープン)
  protected function __construct($dsn, $user, $pass) {
    $this->db = new PDO($dsn, $user, $pass);
    $this->initDb();
  }

  // デストラクタ(DBクローズ)
  function __destruct() {
    unset($this->db);
  }
}


// サービスのメイン(Web APIとデータベースとの連携)
class PalsLine extends DB {

  // コンストラクタ(DBオープン)
  // デストラクタ(DBクローズ)はスーパークラスのものを利用
  public function __construct() {
    parent::__construct('mysql:dbname=palsline;host=localhost', 'root', '');
  }

  // データベース初期化
  protected function initDb() {
    $this->db->exec('SET CHARACTER SET utf8;');
    $this->db->exec('CHARSET utf8;');
  }

  // セッション開始 - ユーザーIDとログイン時刻を記録
  // 戻り値: セッションID
  private function startSession($user_id) {
    $sql = 'INSERT INTO session (user_id) VALUES(?)';
    $this->set($sql, array($user_id));
    $result = $this->lastInsertId;
    return $result;
  }

  // セッション終了 - 現在時刻を記録(ただし有効期限内のみ)
  // 戻り値: 0ならログアウト済み、1なら時刻が記録されログアウト完了
  private function stopSession($session_id) {
    $sql = 'UPDATE session SET logout = NOW() WHERE id = ? AND logout >= NOW()';
    return $this->set($sql, array($session_id));
  }

  // ログアウトまでの時刻を延長
  // 戻り値: 延長できたらtrue、ログアウト済みならfalse
  public function extendSession($session_id, $time = false) {
    $session_id += 0;
    // ログイン中であることを確認
    $sql = 'SELECT id FROM session WHERE id = ? AND login <= NOW()' .
           ' ORDER BY login DESC LIMIT 1';
    $result = $this->get($sql, array($session_id));
    if ($result['id'] + 0 !== $session_id) return false;
    // ログアウトの時刻を更新
    $time = $time ? 'FROM_UNIXTIME(' . ($time + 0) . ')' :
                    'NOW() + INTERVAL 30 MINUTE';
    $sql = "UPDATE session SET logout = $time WHERE id = ?";
    $this->set($sql, array($session_id));
    return true;
  }

  // ユーザー追加(データベースに保存されていないユーザー名が入力されたとき)
  // 戻り値: array('id' => ユーザーID, 'name' => ユーザー名)
  public function addUser($user, $pass) {
    $sql = 'INSERT INTO user(name, pass) VALUES (?, PASSWORD(?));';
    $this->set($sql, array($user, $pass));
    return $this->get('SELECT id, name FROM user WHERE id = ?',
                      array($this->lastInsertId));
  }

  // ログイン - セッション開始
  // 戻り値:   ログイン完了なら連想配列、ログイン不可ならnull
  // 連想配列: array('user' => ユーザー名, 'session_id' => セッションID)
  public function login($user, $pass) {
    // ログインの状態を判定
    $sql = <<<SQL
SELECT
  (SELECT COUNT(uu.id) FROM user uu WHERE uu.name = ?) +
  COUNT(u.id) +
  (SELECT IF(COUNT(s.id) > 0, 1, 0) FROM  logging_session s
                                    WHERE s.user_id = u.id) AS login
  FROM user u WHERE u.name = ? AND u.pass = PASSWORD(?);
SQL;
    $result = $this->get($sql, array($user, $user, $pass));
    switch($result['login']) {
      case 0:   // ログイン可 - 新ユーザー登録
        $user = $this->addUser($user, $pass);
        $result = array('session_id' => $this->startSession($user['id']),
                        'user'       => $user['name']);
        break;
      case 1:   // ログイン不可 - ユーザー名とパスワードが合わない
        $result = null;
        break;
      case 2:   // ログイン可 - 登録済みユーザー
        $user = $this->get('SELECT id, name FROM user WHERE name = ? AND pass = PASSWORD(?);', array($user, $pass));
        $result = array('session_id' => $this->startSession($user['id']),
                        'user' => $user['name']);
        break;
      case 3:   // ログイン済み(セッション継続中)
        $result = $this->get('SELECT ls.id AS session_id, u.name AS user FROM logging_session ls, user u WHERE u.name = ? AND pass = PASSWORD(?) AND u.id = ls.user_id ORDER BY ls.login DESC LIMIT 1;', array($user, $pass));
        $result['session_id'] = $result['session_id'] + 0;
    }
    return $result;
  }

  // ログアウト - セッション終了
  // 戻り値: stopSessionに準じる
  public function logout($session_id) {
    $session_id += 0;
    return $this->stopSession($session_id);
  }

  // メッセージ一覧を参照
  // 戻り値: 参照するデータの配列
  public function getMessages($session_id, $offset = 0, $rows = 10) {
    $session_id += 0;  $offset += 0;  $rows += 0;

    // メッセージ一覧を参照
    $sql = <<<SQL
SELECT u.name AS user, mp.message_id, mp.content,
       (mp.user_id = ls.user_id) owner,
       GROUP_CONCAT(mp.photo_id) AS photos,
       (SELECT COUNT(user_id) - 1 FROM  share
                                  WHERE message_id = mp.message_id) AS share
  FROM message_photo mp, logging_session ls, share s, user u
 WHERE ls.id = ? AND ls.user_id = s.user_id AND
       mp.user_id = u.id AND s.message_id = mp.message_id
 GROUP BY mp.message_id
 ORDER BY mp.regist DESC
 LIMIT $offset, $rows;
SQL;
    $results = $this->get($sql, array($session_id));

    // 1件だけでも配列にする
    if (isset($results['photos'])) $results = array($results);

    // 処理しやすい値に置換
    foreach($results as &$result) {
      // カンマ区切りの写真IDを配列に置換
      if (isset($result['photos'])) {
        $result['photos'] = explode(',', $result['photos']);
      }
      // ownerの値をtrue/falseに置換
      if (isset($result['owner'])) {
        $result['owner'] = ($result['owner'] == 1);
      }
    }
    return $results;
  }

  // メッセージ削除
  // 戻り値: 削除できたら1、削除できなければnull
  public function deleteMessage($session_id, $message_id) {
    $session_id += 0;  $message_id += 0;

    $this->db->beginTransaction();
    try {
      // 削除するメッセージを確認
      $sql = <<<SQL
SELECT m.id FROM message m, logging_session ls
 WHERE ls.id = ? AND m.id = ? AND m.user_id = ls.user_id;
SQL;
      $delete = $this->get($sql, array($session_id, $message_id));
      if (!isset($delete) || !isset($delete['id'])) return;

      // 写真を削除
      $sql = 'DELETE FROM photo WHERE message_id = ?';
      $this->set($sql, array($delete['id']));
      // 共有ユーザー設定を削除
      $sql = 'DELETE FROM share WHERE message_id = ?';
      $this->set($sql, array($delete['id']));
      // メッセージを削除
      $sql = 'DELETE FROM message WHERE id = ?';
      $result = $this->set($sql, array($delete['id']));

      $this->db->commit();    // コミット(削除を確定)
      return $result;
    } catch(PDOException $e) {
      $this->db->rollback();  // ロールバック(削除を取消)
    }
  }

  // 写真一覧を参照
  // 戻り値: 参照するデータの配列
  public function getPhotos($session_id, $offset = 0, $rows = 10) {
    $session_id += 0;  $offset += 0;  $rows += 0;

    // 写真データを参照
    $sql = <<<SQL
SELECT u.name AS user, mp.photo_id AS id, mp.message_id,
       (mp.user_id = ls.user_id) AS owner,
       (SELECT COUNT(user_id) - 1 FROM  share
                                  WHERE message_id = mp.message_id) AS share
  FROM message_photo mp, share s, logging_session ls, user u
 WHERE ls.id = ? AND s.user_id = ls.user_id AND
       mp.photo_id IS NOT NULL AND mp.message_id = s.message_id AND
       mp.user_id = u.id
 ORDER BY mp.regist desc, mp.photo_id desc
 LIMIT $offset, $rows;
SQL;
    $results = $this->get($sql, array($session_id));

    // 1件でも配列にする
    if (isset($results['user'])) $results = array($results);

    // ownerの値を0/1からfalse/trueに置換
    foreach($results as &$result) {
      if (!isset($result['owner']))  continue;
      $result['owner'] = ($result['owner'] == 1);
    }
    return $results;
  }

  // 写真データを参照(<img>用)
  // 戻り値: array('type' => MIMEタイプ, 'data' => 写真データ)
  public function getImage($session_id, $photo_id) {
    $session_id += 0;  $photo_id += 0;

    $sql = <<<SQL
SELECT mp.type, mp.data
  FROM message_photo mp, logging_session ls, share s
 WHERE ls.id = ? AND mp.photo_id = ? AND
       mp.message_id = s.message_id AND ls.user_id = s.user_id
 ORDER BY mp.regist;
SQL;
    return $this->get($sql, array($session_id, $photo_id));
  }

  // 写真削除
  // 戻り値: 削除できたら1、削除できなければ0
  public function deletePhoto($session_id, $photo_id) {
    $sql = <<<SQL
SELECT mp.photo_id FROM message_photo mp, logging_session ls
 WHERE ls.id = ? AND mp.photo_id = ? AND mp.user_id = ls.user_id;
SQL;
    $result = $this->get($sql, array($session_id, $photo_id));
    return $this->set('DELETE FROM photo WHERE id = ?',
                      array($result['photo_id']));
  }

  // ユーザー一覧を参照(メッセージ投稿用)
  // 戻り値: ユーザー一覧データの配列
  public function getUsers($session_id) {
    $session_id += 0;
    // ログインしている以外のユーザー一覧を参照
    $sql = 'SELECT id, name, 0 AS shared FROM user u' .
           ' WHERE id <> (SELECT user_id FROM logging_session ls' .
           ' WHERE ls.id = ? ORDER BY login DESC LIMIT 1);';
    $results = $this->get($sql, array($session_id));
    if (isset($results['id'])) $results = array($results);
    // 処理しやすい値に置換
    foreach($results as &$result) {
      $result['shared'] = ($result['shared'] == 1);
    }
    return $results;
  }

  // メッセージ追加(投稿) - メッセージ・写真・共有ユーザー設定
  // 戻り値: メッセージID
  public function addMessage($session_id, $message,
                             $share = false, $files = false) {
    $session_id += 0;

    $this->db->beginTransaction();  // トランザクション開始
    try {
      // メッセージ追加
      $sql = 'INSERT INTO message (user_id, content)' .
             ' SELECT user_id, ? FROM logging_session WHERE id = ?';
      $this->set($sql, array($message, $session_id));
      $message_id = $this->lastInsertId;

      // 投稿ユーザー設定
      $sql = 'INSERT INTO share (message_id, user_id)' .
             ' SELECT ?, user_id FROM logging_session WHERE id = ?';
      $this->set($sql, array($message_id, $session_id));

      // 共有ユーザー設定
      if (isset($message_id) && $share) {
        $sql = 'INSERT INTO share (message_id, user_id)' .
               ' SELECT ?, id FROM user WHERE id = ?';
        foreach($share as $user_id) {
          $this->set($sql, array($message_id, $user_id));
        }
      }

      // 写真データ追加
      if (isset($message_id) && $files) {
        $sql = 'INSERT INTO photo (message_id, name, type, data)' .
               ' VALUES (?, ?, ?, ?) ON DUPLICATE KEY UPDATE' .
               ' type = VALUES(type), data = VALUES(data);';
        $len = count($files['name']);
        for($i = 0; $i < $len; $i++) {
          if (filesize($files['tmp_name'][$i]) < 1) continue;
          $this->set($sql, array($message_id, $files['name'][$i],
                                 $files['type'][$i],
                                 file_get_contents($files['tmp_name'][$i])));
        }
      }

      $this->db->commit();  // コミット(追加を確定)

    } catch(PDOException $e) {
      $this->db->rollback();  // ロールバック(追加を取消)
      $message_id = null;
    }

    return $message_id;
  }

  // メッセージ参照
  // 戻り値: 参照するメッセージに関するデータ(連想配列)
  public function getMessage($session_id, $message_id) {
    $session_id += 0;  $message_id += 0;

    // 指定されたメッセージの参照
    $sql = <<<SQL
SELECT u.name AS user, mp.message_id, mp.content,
       GROUP_CONCAT(mp.photo_id) AS photos,
       COUNT(DISTINCT s.user_id) - 1 AS share,
       (mp.user_id = ls.user_id) owner,
       (SELECT MAX(pm.id) FROM message pm, share ps
         WHERE pm.id < mp.message_id AND
               pm.id = ps.message_id AND ps.user_id = ls.user_id) AS prev,
       (SELECT MIN(nm.id) FROM message nm, share ns
         WHERE nm.id > mp.message_id AND
               nm.id = ns.message_id AND ns.user_id = ls.user_id) AS next
  FROM message_photo mp, logging_session ls, share s, user u
 WHERE ls.id = ? AND mp.message_id = ? AND ls.user_id = s.user_id AND
       s.message_id = mp.message_id AND mp.user_id = u.id
 GROUP BY mp.message_id;
SQL;
    $result = $this->get($sql, array($session_id, $message_id));

    // owner, photosのデータを置換
    // owner: 0/1からfalse/trueへ, photos: カンマ区切りから配列へ
    $result['owner']  = ($result['owner'] == 1);
    if (isset($result['photos'])) {
      $result['photos'] = explode(',', $result['photos']);
    }
    return $result;
  }

  // メッセージ共有ユーザー一覧を参照(メッセージ表示用)
  // 戻り値: ユーザー一覧データの配列
  public function getShared($session_id, $message_id) {
    $session_id += 0;  $message_id += 0;

    // メッセージ共有ユーザーの一覧を参照
    // (共有済みのユーザーとそうでないユーザーとを区別する)
    $sql = <<<SQL
SELECT u.name, mu.user_id AS id, s.user_id IS NOT NULL AS shared
  FROM (SELECT m.id message_id, u.id AS user_id FROM message m, user u) mu
       LEFT OUTER JOIN
       share s
       ON mu.message_id = s.message_id AND mu.user_id = s.user_id,
       user u, logging_session ls
 WHERE ls.id = ? AND mu.message_id = ? AND
       mu.user_id = u.id AND ls.user_id <> mu.user_id;
SQL;

    $results = $this->get($sql, array($session_id, $message_id));
    // 1件でも配列にする
    if (isset($results['id'])) $results = array($results);
    // sharedの値を0/1からfalse/trueに変換
    foreach($results as &$result) {
      $result['shared'] = $result['shared'] == 1;
    }

    return $results;
  }

  // メッセージ共有ユーザーを設定
  // 戻り値: メッセージを共有したユーザーの数
  public function addShare($session_id, $message_id, $share) {
    $session_id += 0;  $message_id += 0;

    $result = 0;
    $this->db->beginTransaction();
    try {
      // それまでの設定を削除
      $sql = <<<SQL
DELETE FROM share
 WHERE message_id = ? AND
       user_id <> (SELECT user_id FROM logging_session WHERE id = ?);
SQL;
      $this->set($sql, array($message_id, $session_id));

      // 共有ユーザーを設定
      $sql = <<<SQL
INSERT INTO share (message_id, user_id)
       SELECT m.id message_id, u.id user_id
         FROM message m, logging_session ls, user u
        WHERE ls.id = ? AND m.id = ? AND u.id = ? AND
              ls.user_id = m.user_id;
SQL;
      foreach($share as $user_id) {
        $result += $this->set($sql, array($session_id, $message_id, $user_id));
      }

      $this->db->commit();    // コミット(設定を確定)
    } catch(PDOException $e) {
      $this->db->rollback();  // ロールバック(設定を取消)
      $result = 0;
    }
    return $result;
  }
}
